/******************************************************************************
 * FilterWheel.ino
 *
 * 5-position color filter wheel
 *
 * Inputs:
 *
 * 3 Hall effect switches on A0-A2
 *
 * Ouputs:
 *
 * STEP on pin D2
 * DIR on pin D3
 * EN on pin D4
 *
 * On power up it looks for filter 1 and centers it.
 *
 * The motor makes ~15.76 turns per turn of the wheel, or 6304 steps per turn.
 *****************************************************************************/

#include <AccelStepper.h>

#define MAX_SLEW_SPEED 4000
#define MAX_SLEW_ACCEL 40000

// The motor instance.
AccelStepper motor(AccelStepper::DRIVER, 2, 3);

// Motor enable pin.
uint8_t enablePin = 4;

// The currently selected filter.
uint8_t currentFilter = 0;

// How far to run when skipping 0, 1, 2, 3, or 4 filters.
long positions[5] = {0, 1262, 2526, 3788, 5048};

// A fudge factor to drive past the filter while reading the
// filter index edges.
long overshoot = 150;

// Reset vector so we can do a software reset.
void (*reset)(void) = 0;

/******************************************************************************
 * findFilterTrailingEdge()
 *
 * Finds the trailing edge of the first filter it sees.
 *
 *****************************************************************************/

void findFilterTrailingEdge() {

    // If the wheel is stopped in between filters,
    // run to the next filter.
    if (currentFilter == 0) {
        motor.move(positions[1]);
        while (motor.run() && (currentFilter == 0));
    }

    // Run to the trailing edge of the filter index.
    motor.move(positions[1]);
    while (motor.run() && (currentFilter != 0));
    motor.stop();
    while (motor.run());
}

/******************************************************************************
 * findAnyFilter()
 *
 * Finds the center of the first filter it sees.
 *
 *****************************************************************************/

void findAnyFilter() {

    int distance = 0;
    long steps = 0;
    long begin = 0;
    long end = 0;
    long center = 0;

    // Are we on a filter?
    if (currentFilter != 0) {
        // Go find the trailing edge.
        findFilterTrailingEdge();
    }
    // Look (one filter spacing + fudge) ahead for a
    // filter index to start.
    steps = positions[1] + overshoot;

    motor.move(steps);
    while (currentFilter == 0) {
        motor.run();
    }
    begin = motor.currentPosition();

    // Find the end of the filter index, then go to
    // the center of the index.
    while (currentFilter != 0) {
        motor.run();
    }
    end = motor.currentPosition();
    motor.stop();
    while(motor.run());

    center = (end + begin) >> 1;
    motor.moveTo(center);
    while(motor.run());
}

/******************************************************************************
 * runToCenter(int filter)
 *
 * Runs to the specified filter and centers.
 *
 *****************************************************************************/

void runToCenter(int filterNo) {

    int distance = 0;
    long steps = 0;
    long begin = 0;
    long end = 0;
    long center = 0;

    // Do we need to move at all?
    if (currentFilter != filterNo) {

        // If not on a filter, find one and center it.
        if (currentFilter == 0)
            findAnyFilter();

        // Figure out how far (in filters) to the target.
        if (filterNo > currentFilter) {
            distance = filterNo - currentFilter;
        } else {
            distance = 5 - currentFilter + filterNo;
        }
        // Convert to steps and add the fudge.
        steps = positions[distance];
        steps += overshoot;

        // Run to within 250 steps of the target.
        motor.setCurrentPosition(0);
        motor.move(steps);
        while (motor.distanceToGo() > 250) {
            motor.run();
        }

        // Run until we hit a filter index and
        // mark the position.
        while (currentFilter == 0) {
            motor.run();
        }
        begin = motor.currentPosition();

        // Run until we are past the filter index and
        // mark the position.
        while (currentFilter != 0) {
            motor.run();
        }
        end = motor.currentPosition();

        // Stop the motor, and back up to the center of
        // the filter index.
        motor.stop();
        while(motor.run());
        delay(25);
        center = (end + begin) >> 1;
        motor.moveTo(center);
        while(motor.run());
    }
}

/******************************************************************************
 * findFilter1()
 *
 * Finds the center of filter1.
 *
 *****************************************************************************/

void findFilter1() {

    // Get to the center of a filter - any filter, so
    // we know where we are.
    findAnyFilter();

    // Go to the center of filter 1.
    runToCenter(1);
}

/******************************************************************************
 *
 * pciSetup()
 *
 * Setup pins for pin change interrupt.
 *
 *****************************************************************************/

void pciSetup(uint8_t pin) {

    *digitalPinToPCMSK(pin) |= bit(digitalPinToPCMSKbit(pin));
    PCIFR |= bit(digitalPinToPCICRbit(pin)); // clear any outstanding interrupt
    PCICR |= bit(digitalPinToPCICRbit(pin)); // enable interrupt for the group
}

/******************************************************************************
 *
 * setup()
 *
 *****************************************************************************/

void setup() {

    // Start the camera interface.
    Serial.begin(9600);

    // Enable pull-ups on Hall switch inputs.
    digitalWrite(A0, HIGH);
    digitalWrite(A1, HIGH);
    digitalWrite(A2, HIGH);

    // Enable the pin change interrupts.
    pciSetup(A0);
    pciSetup(A1);
    pciSetup(A2);

    // Setup the motor.
    motor.setEnablePin(enablePin);
    motor.setMaxSpeed(MAX_SLEW_SPEED);
    motor.setAcceleration(MAX_SLEW_ACCEL);
    motor.setCurrentPosition(0);
    motor.setPinsInverted(true, false, true);
    currentFilter = (PINC & 0x07) ^ 0x07;

    // Go to filter 1
    motor.enableOutputs();
    findFilter1();
    delay(10);
    motor.disableOutputs();
}

/******************************************************************************
 *
 * loop()
 *
 * Reads one ASCII character ('0' to '4') from the camera and converts it to a
 * binary 1 - 5 value to match the magnets on the filters.
 *
 *****************************************************************************/

void loop() {

    int ch = Serial.read();

    // Is it a filter?
    if ((ch >= '0') && (ch <= '4')) {
        ch -= '0';
        motor.enableOutputs();
        runToCenter(ch + 1);
        delay(10);
        motor.disableOutputs();
    }
    // Is it a software reset request?
    else if (ch == 'r') {
        reset();
    }
}

/******************************************************************************
 *
 * Pin change interrupt for the Hall switches.
 *
 * Sets the current filter to the value of the magnetic switches.
 *
 *****************************************************************************/

ISR (PCINT1_vect) {
    currentFilter = (PINC & 0x07) ^ 0x07;
}

